import { isFSA } from 'flux-standard-action';
import qs from 'qs';
import createBrowserHistory from 'history/createBrowserHistory';
import createMemoryHistory from 'history/createMemoryHistory';
import createHashHistory from 'history/createHashHistory';

export const ACTION_PREFIX = '@@I_REDUX_NAVIGATION';

export const HISTORY_TYPES = {
  BROWSER: Symbol('browser'),
  HASH: Symbol('hash'),
  MEMORY: Symbol('memory'),
};

const trimSearchStr = (search = '') => search.replace(/^\?/, '');

export const createHistory = (type = HISTORY_TYPES.BROWSER, params = {}) => {
  const {
    basename = '',
    forceRefresh = false,
    keyLength = 6,
    hashType = 'noslash',
    initialEntries = [],
    initialIndex = 0,
    getUserConfirmation = null,
  } = params;

  let historyCreator;
  let historyParams = { getUserConfirmation: null };

  switch (type) {
    case HISTORY_TYPES.BROWSER:
      historyParams = { basename, forceRefresh, keyLength, getUserConfirmation };
      historyCreator = createBrowserHistory;
      break;
    case HISTORY_TYPES.HASH:
      historyParams = { basename, hashType, getUserConfirmation };
      historyCreator = createHashHistory;
      break;
    case HISTORY_TYPES.MEMORY:
      historyParams = { initialEntries, initialIndex, keyLength, getUserConfirmation };
      historyCreator = createMemoryHistory;
      break;
    default:
      throw Error(`The given type of history "${type}" is not supported yet.`);
  }

  return historyCreator(historyParams);
};

export const createMiddleware = (history, { actionPrefix = ACTION_PREFIX }) => ({ dispatch }) => {
  history.listen((location, action) => {
    dispatch({
      type: `${actionPrefix}/CHANGE`,
      payload: {
        location: { ...location, query: qs.parse(trimSearchStr(location.search), { parameterLimit: 2000, arrayLimit: 2000 }) },
        action
      }
    });
  });

  return (next) => (action) => {
    if (isFSA(action) && action.type.indexOf(actionPrefix) === 0) {
      const { type, payload: args } = action;
      const historyAction = type.split('/')[1] || '';

      switch (historyAction) {
        case 'PUSH':
          history.push(...args);
          break;
        case 'REPLACE':
          history.replace(...args);
          break;
        case 'GO':
          history.go(...args);
          break;
        case 'GO_BACK':
          history.goBack(...args);
          break;
        case 'GO_FORWARD':
          history.goForward(...args);
          break;
        case 'CAN_GO':
          history.canGo(...args);
          break;
      }
    }

    return next(action);
  };
};

export const createReducer = (initialState = {}, { actionPrefix = ACTION_PREFIX }) => (state = initialState, action = {}) => {
  if (isFSA(action) && action.type === `${actionPrefix}/CHANGE`) {
    const { location } = action.payload;
    location.query = location.query || {};

    state = { ...state, location };
  }

  return state;
};

export const createActions = ({ actionPrefix = ACTION_PREFIX }) => {
  const push = (...args) => ({ type: `${actionPrefix}/PUSH`, payload: args });
  const replace = (...args) => ({ type: `${actionPrefix}/REPLACE`, payload: args });
  const go = (...args) => ({ type: `${actionPrefix}/GO`, payload: args });
  const goBack = (...args) => ({ type: `${actionPrefix}/GO_BACK`, payload: args });
  const goForward = (...args) => ({ type: `${actionPrefix}/GO_FORWARD`, payload: args });
  const canGo = (...args) => ({ type: `${actionPrefix}/CAN_GO`, payload: args });

  return { push, replace, go, goBack, goForward, canGo };
};

export const createNavigation = (params = {}) => {
  const {
    actionPrefix = ACTION_PREFIX,
    historyType = HISTORY_TYPES.BROWSER,
    historyParams = {},
  } = params;

  const { history = createHistory(historyType, historyParams) } = params;
  const initialReducerState = {
    location: {
      ...history.location,
      query: history.location.query || qs.parse(trimSearchStr(history.location.search), { parameterLimit: 2000, arrayLimit: 2000 })
    }
  };

  const {
    middleware = createMiddleware(history, { actionPrefix }),
    reducer = createReducer(initialReducerState, { actionPrefix }),
    actions = createActions({ actionPrefix })
  } = params;

  return { reducer, history, middleware, actions };
};

export default createNavigation;
